/*
** Copyright (c) 2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: mod_misc.c,v 1.4 2004/08/12 10:28:49 bambi Exp $
**
*/

/*
** Module	: Standard Library Module
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>

#include <config/config.h>

#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif

#include "ember.h"
#include "redblack.h"



/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/


/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/


/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/


int doTypeOf(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*typeName;

	if (!args)
	{
		strcpy(script->errorMsg, "Too few parameters to typeof()");
		return(-1);
	}
	if (args->next)
	{
		strcpy(script->errorMsg,"Too many parameters to typeof()");
		return(-1);
	}
	switch(args->value.type)
	{
		case EMBER_TYPE_TEXT:
			typeName = "text";
			break;

		case EMBER_TYPE_INT:
			typeName = "int";
			break;

		case EMBER_TYPE_REAL:
			typeName = "real";
			break;

		case EMBER_TYPE_UINT:
			typeName = "uint";
			break;

		case EMBER_TYPE_MONEY:
			typeName = "money";
			break;

		case EMBER_TYPE_ARRAY:
			typeName = "array";
			break;

		case EMBER_TYPE_ASSOC:
			typeName = "assoc";
			break;

		default:
			typeName = "unknown";
			break;
	}
	eSymSetTextValue(script,&script->returnValue, typeName);
        return(0);
}


int doCreateVariable(script, args)
	ember	*script;
	e_stack	*args;
{
	char	*name,
		*value;
	e_stab	*sym;

	if (!args)
	{
		strcpy(script->errorMsg, "Too few parameters");
		return(-1);
	}
	name = emTextArg(args);
	value = emTextArg(args->next);

	sym = eSymGetVariableByName(script, name);
	if (sym == NULL)
		sym = eSymCreateVariable(script, name);
	if (sym == NULL)
		return(-1);
        eSymSetTextValue(script, &sym->value, value);
	return(0);
}


int doKeys(script, args)
	ember	*script;
	e_stack	*args;
{
	rbtree	*tree;
        e_assoc *curElement;
	e_value val;
        int     count;

	if (!args)
	{
		strcpy(script->errorMsg, "Too few parameters");
		return(-1);
	}
	if (args->value.type != EMBER_TYPE_ASSOC)
	{
		strcpy(script->errorMsg, "Variable not an assoc array");
		return(-1);
	}

	val.initialised = EMBER_FALSE;
        count = 0;
	tree = (rbtree*)args->value.value.assocVal;
	curElement = (e_assoc*)rblookup(RB_LUFIRST, NULL, tree);
	if (curElement->value.initialised == EMBER_FALSE)
	{
		return(0);
	}
        while(curElement)
        {
		eSymSetTextValue(script, &val, curElement->index);
		eSymSetArrayElement(script, &script->returnValue,
                                &val, count);
		eSymFreeValue(script, &val);
                count++;
		curElement = (e_assoc*)rblookup(RB_LUNEXT, curElement, tree);
        }
	return(0);
}


