%{
/*
** Copyright (c) 2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  
**
** Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: parser.y,v 1.26 2007/07/16 06:12:15 bambi Exp $
**
*/

/*
** Module	: 
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/



/**************************************************************************
** STANDARD INCLUDES
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>

#include <config/config.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif


/**************************************************************************
** MODULE SPECIFIC INCLUDES
**************************************************************************/

#include "ember.h"

int	yylex();
void	yyerror();

extern	ember *emberScript;


/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/


/* Generic stack.  Used for tracking things within the parser */

typedef struct gen_stack_frame_s {
	void	*ptr;
	struct gen_stack_frame_s *next;
} gen_stack_frame;


static gen_stack_frame *generic_stack = NULL;

void genericStackPush(ptr)
	void	*ptr;
{
	gen_stack_frame	*cur;

	cur = (gen_stack_frame *)malloc(sizeof(gen_stack_frame));
	cur->ptr = ptr;
	cur->next = generic_stack;
	generic_stack = cur;
}

void *genericStackPop()
{
	gen_stack_frame		*cur;
	void			*ptr;

	if (!generic_stack)
	{
		return(NULL);
	}
	cur = generic_stack;
	ptr = cur->ptr;
	generic_stack = generic_stack->next;
	free(cur);
	return(ptr);
}

/**************************************************************************
** YACC DEFINITION
**************************************************************************/

#define YYDEBUG 1

%}

%token	EXEC_PATH

%token  HTML
%token  HTML_START
%token  HTML_END

%token  COND_GE
%token  COND_LE
%token  COND_NE
%token  COND_EQ  
%token  COND_GT
%token  COND_LT
%token  COND_RE
%token  COND_NRE
%token	COND_AND
%token	COND_OR

%token  OP_AUTOADD 
%token  OP_AUTOSUB
%token  OP_AUTOMUL
%token  OP_AUTODIV
%token  OP_INC
%token  OP_DEC
%token  OP_SHIFT_L
%token  OP_SHIFT_R

%token	OP_ADD
%token	OP_SUB
%token	OP_MUL
%token	OP_DIV
%token	OP_MOD
%token	OP_AND
%token	OP_OR

%token  IF
%token  ELSE
%token  WHILE
%token  BREAK
%token  CONTINUE
%token  RETURN
%token  FUNCT
%token  COUNT
%token  DEFINED
%token  UNDEF
%token  LOAD
%token  MOD_LOAD

%token  PUSH
%token  POP
%token  SHIFT
%token  UNSHIFT

%token	L_BRACE
%token	R_BRACE
%token	L_BRACKET
%token	R_BRACKET
%token	L_SQUARE
%token	R_SQUARE
%token	ASSIGN
%token	BANG
%token	COMMA
%token	SEMI_COLON

%token  T_INT
%token  T_UINT
%token  T_TEXT
%token  T_CHAR
%token  T_REAL
%token  T_MONEY
%token  T_ARRAY
%token  T_ASSOC


%token  VAR
%token  IDENT   

%token  LIT_TEXT
%token  LIT_INT
%token  LIT_REAL
%token  LIT_NEG_INT
%token  LIT_NEG_REAL

%token	BAD_LEX

%left   OP_ADD OP_SUB OP_INC OP_DEC
%left   OP_MUL OP_DIV OP_MOD OP_AND OP_OR
%left   SHIFT_LEFT SHIFT_RIGHT
%left   COUNT
%right	ASSIGN OP_AUTOADD

%%

ember_script
	: opt_exec_path
	  {
		eInit(emberScript);
	  }
	  opt_load
	  function_definitions
	  {
		eInitMain(emberScript);
	  }
	  main_definitions
	  {
		eEndOfInput();
	  }
	;

opt_exec_path
	: /* NULL */
	| EXEC_PATH
	;

opt_load
	: /* NULL */
	| html_content opt_load
	| LOAD LIT_TEXT SEMI_COLON 
	  {
		if (eLoadLibrary(emberScript, $2) < 0)
		{
			free($2);
			return(-1);
		}
		free($2);
	  } opt_load
	| MOD_LOAD LIT_TEXT SEMI_COLON 
	  {
		if (eDynLoadModule(emberScript, $2) < 0)
		{
			free($2);
			return(-1);
		}
		free($2);
	  } opt_load
	  ;

function_definitions
	: /* NULL */
	| function_def L_BRACE ember_code retval  R_BRACE function_definitions
	| html_content function_definitions
	;

function_def
	: FUNCT IDENT 
	  {
		if (eSymCreateFunction(emberScript, $2) < 0)
		{
			free($2);
			return(-1);
		}
		free($2);
	  } 
	  L_BRACKET opt_function_args R_BRACKET
	  ;

opt_function_args
	: /* NULL */
	| function_arg opt_function_tail
	;

opt_function_tail
	: /* NULL */
	| COMMA function_arg opt_function_tail
	;

function_arg
	: type VAR
	{
		if (eSymAddFunctionArg(emberScript, $2, (long)$1) < 0)
		{
			free($2);
			return(-1);
		}
		free($2);
	}
	;

retval
	: /* NULL */
	  {
		/*
		** Ensure that the last opcode we wrote was a return.  If
		** not then we'd better put one in there so that this
		** thing has a return value (even a default empty value)
		*/
		if (emberScript->info.curFunction->lastPcode == NULL ||
		    emberScript->info.curFunction->lastPcode->op != E_RETURN)
		{
			if (eCodePushLiteral(emberScript, EMBER_TYPE_TEXT, 
				"") < 0)
			{
				return(-1);
			}
		}
	  }
	  ;

main_definitions
	: /* NULL */
	| 
	  {
		emberScript->info.curFunction = &emberScript->main;
	  }
	ember_code
	;



/*
**  Main script code definition 
*/

ember_code
	: /* NULL */
	| html_content ember_code
	| variable_assign SEMI_COLON ember_code
	| { emberScript->info.ignoreReturn = EMBER_TRUE; }
	  function_call SEMI_COLON
	  { emberScript->info.ignoreReturn = EMBER_FALSE; }
	  ember_code
	| if_code_block ember_code
	| while_code_block ember_code
	| return SEMI_COLON ember_code 
	| BREAK SEMI_COLON 
	  {
		int	label;

		label = eCodePopBreakLabel(emberScript);
		eCodeJMP(emberScript, label);
		eCodePushBreakLabel(emberScript, label);
	   }
	  ember_code
	| CONTINUE SEMI_COLON 
	  {
		int	label;

		label = eCodePopContLabel(emberScript);
		eCodeJMPB(emberScript, label);
		eCodePushContLabel(emberScript, label);
	   }
	  ember_code

	| array_op SEMI_COLON ember_code
	  ;
	| misc_op SEMI_COLON ember_code
	  ;

return
	: RETURN L_BRACKET opt_return_value R_BRACKET
	  {
		if (eCodeReturn(emberScript) < 0)
			return(-1);
	  }
	  ;

opt_return_value
	: /* NULL */
	  {
		if (eCodePushLiteral(emberScript, EMBER_TYPE_TEXT, "") < 0)
		{
			return(-1);
		}
	  }
	| expr
	;


variable_assign
	: VAR ASSIGN expr 
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$1);
		if (symbol == NULL)
			symbol = eSymCreateVariable(emberScript,$1);
		if (symbol == NULL)
		{
			free($1);
			return(-1);
		}
		varID = symbol->id;
		if (eCodeStore(emberScript, varID) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }

	| VAR L_SQUARE expr R_SQUARE ASSIGN expr 
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$1);
		if (symbol == NULL)
			symbol = eSymCreateVariable(emberScript,$1);
		if (symbol == NULL)
		{
			free($1);
			return(-1);
		}
		varID = symbol->id;
		if (eCodeArrayStore(emberScript, varID) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }

	| VAR L_BRACE expr R_BRACE ASSIGN expr 
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$1);
		if (symbol == NULL)
			symbol = eSymCreateVariable(emberScript,$1);
		if (symbol == NULL)
		{
			free($1);
			return(-1);
		}
		varID = symbol->id;
		if (eCodeAssocStore(emberScript, varID) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }

	| VAR OP_INC
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$1);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$1);
		}
		varID = symbol->id;
		if (eCodePushVariable(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodePushLiteral(emberScript, EMBER_TYPE_INT, "1") < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeAlu(emberScript, EMBER_OP_ADD) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeStore(emberScript, varID) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }

	| VAR OP_DEC
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$1);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$1);
		}
		varID = symbol->id;
		if (eCodePushVariable(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodePushLiteral(emberScript, EMBER_TYPE_INT, "1") < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeAlu(emberScript, EMBER_OP_SUB) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeStore(emberScript, varID) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
/*
** AUTOSUB and AUTODIV are broken as the ALU args are pushed in the wrong
** order.  Do we really need auto-ops anyway?   Commented out.
***
	| VAR OP_AUTOADD
	  {
		e_stab	*symbol;
		void	*ptr;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$1);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$1);
		}
		varID = symbol->id;
		if (eCodePushVariable(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
		ptr = malloc(sizeof(varID));
		bcopy(&varID, ptr, sizeof(varID));
		genericStackPush(ptr);
	  }
	  expr
	  {
		void	*ptr;
		int	varID;

		ptr = (void*)genericStackPop();
		bcopy(ptr, &varID, sizeof(varID));
		
		if (eCodeAlu(emberScript, EMBER_OP_ADD) < 0)
		{
			return(-1);
		}
		if (eCodeStore(emberScript, varID) < 0)
		{
			return(-1);
		}
	}

	| VAR OP_AUTOSUB expr
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$1);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$1);
		}
		varID = symbol->id;
		if (eCodePushVariable(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeAlu(emberScript, EMBER_OP_SUB) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeStore(emberScript, varID) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	}

	| VAR OP_AUTOMUL expr
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$1);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$1);
		}
		varID = symbol->id;
		if (eCodePushVariable(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeAlu(emberScript, EMBER_OP_MUL) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeStore(emberScript, varID) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	}
**
*/
	;


function_call
	: IDENT L_BRACKET
	  {
		if (eCodeBlockIn(emberScript) < 0)
		{
			free($1);
			return(-1);
		}
	  }
	  opt_param_list R_BRACKET
	  {
		if (eCodeCall(emberScript, $1, EMBER_TRUE) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeBlockOut(emberScript) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	  ;

opt_param_list
	: /* NULL */
	| expr opt_param_tail
	;

opt_param_tail
	: /* NULL */
	| COMMA expr opt_param_tail
	;


if_code_block
	: {
		int	startLabel,
			elseLabel,
			endLabel;

		startLabel = eCodeGetLabel(emberScript);
		elseLabel = eCodeGetLabel(emberScript);
		endLabel = eCodeGetLabel(emberScript);
		eCodePushLabel(emberScript,endLabel);
		eCodePushLabel(emberScript,elseLabel);
		eCodePushLabel(emberScript,startLabel);
	  }
	IF L_BRACKET condition R_BRACKET
	  {
		int	startLabel,
			elseLabel,
			endLabel;

		startLabel = eCodePopLabel(emberScript);
		elseLabel = eCodePopLabel(emberScript);
		endLabel = eCodePopLabel(emberScript);
		eCodeJZ(emberScript,elseLabel);
		eCodePushLabel(emberScript,endLabel);
		eCodePushLabel(emberScript,elseLabel);
		eCodeLabel(emberScript,startLabel);
	  }
	  L_BRACE ember_code R_BRACE
	  {
		int	elseLabel,
			endLabel;

		elseLabel = eCodePopLabel(emberScript);
		endLabel = elseLabel + 1;
		eCodeJMP(emberScript,endLabel);
		eCodeLabel(emberScript,elseLabel);
	  }
	  opt_else 
	  {
		int	endLabel;

		endLabel = eCodePopLabel(emberScript);
		eCodeLabel(emberScript,endLabel);
	  }
	  ;


while_code_block
	: {
		int	startLabel,
			endLabel;

		startLabel = eCodeGetLabel(emberScript);
		endLabel = eCodeGetLabel(emberScript);
		eCodeLabel(emberScript,startLabel);
		eCodePushLabel(emberScript, startLabel);
		eCodePushContLabel(emberScript, startLabel);
		eCodePushLabel(emberScript, endLabel);
	  }
	  WHILE L_BRACKET condition R_BRACKET
	  {
		int	endLabel;

		endLabel = eCodePopLabel(emberScript);
		eCodeJZ(emberScript, endLabel);
		eCodePushLabel(emberScript, endLabel);
		eCodePushBreakLabel(emberScript, endLabel);
	  }
	  L_BRACE ember_code R_BRACE
	  {
		int	startLabel,
			endLabel;

		endLabel = eCodePopLabel(emberScript);
		startLabel = eCodePopLabel(emberScript);
		eCodeJMPB(emberScript, startLabel);
		eCodeLabel(emberScript, endLabel);
		eCodePopBreakLabel(emberScript);
		eCodePopContLabel(emberScript);
	  }
	  ;

condition
	: cond_expr opt_cond_logic
	| L_BRACKET condition R_BRACKET opt_cond_logic
	;

cond_expr
	: expr cond_op expr 
	  {
		if (eCodeCMP(emberScript, (long)$2) < 0)
			return(-1);
	  }
	| expr
	  {
		if (eCodePushLiteral(emberScript, EMBER_TYPE_INT, "0") < 0)
			return(-1);
		if (eCodeCMP(emberScript, EMBER_COND_NE) < 0)
			return(-1);
	  }
	| BANG expr
	  {
		if (eCodePushLiteral(emberScript, EMBER_TYPE_INT, "0") < 0)
			return(-1);
		if (eCodeCMP(emberScript, EMBER_COND_EQ) < 0)
			return(-1);
	  }
	  ;


opt_cond_logic
	: /* NULL */
	| {
		int	startLabel,
			elseLabel;

		startLabel = eCodePopLabel(emberScript);
		elseLabel = eCodePopLabel(emberScript);
		if (startLabel < 0 || elseLabel < 0)
			return(-1);
		if (eCodeCondShortcut(emberScript, EMBER_OP_OR, startLabel) < 0)
			return(-1);
		eCodePushLabel(emberScript,elseLabel);
		eCodePushLabel(emberScript,startLabel);
	  }
	  COND_OR condition
	  {
		if (eCodeAlu(emberScript, EMBER_OP_OR) < 0)
			return(-1);
	  }
	| 
	  {
		int	startLabel,
			elseLabel;

		startLabel = eCodePopLabel(emberScript);
		elseLabel = eCodePopLabel(emberScript);
		if (startLabel < 0 || elseLabel < 0)
			return(-1);
		if (eCodeCondShortcut(emberScript, EMBER_OP_AND, elseLabel) < 0)
			return(-1);
		eCodePushLabel(emberScript,elseLabel);
		eCodePushLabel(emberScript,startLabel);
	  }
	  COND_AND condition
	  {
		if (eCodeAlu(emberScript, EMBER_OP_AND) < 0)
			return(-1);
	  }
	;



opt_else
	: /* NULL */
	| ELSE L_BRACE ember_code R_BRACE
	;


array_op
	: PUSH L_BRACKET VAR COMMA expr R_BRACKET
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$3);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$3);
		}
		free($3);
		if (symbol == NULL)
		{
			return(-1);
		}
		varID = symbol->id;
		if (eCodeArrayPush(emberScript, varID) < 0)
		{
			return(-1);
		}
	  }
	| UNSHIFT L_BRACKET VAR COMMA expr R_BRACKET
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$3);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$3);
		}
		free($3);
		if (symbol == NULL)
		{
			return(-1);
		}
		varID = symbol->id;
		if (eCodeArrayUnshift(emberScript, varID) < 0)
		{
			return(-1);
		}
	  }

misc_op
	: UNDEF L_BRACKET VAR R_BRACKET
	  {
		if (eCodeUndef(emberScript, $3) < 0)
		{
			free($3);
			return(-1);
		}
		free($3);
	  }

expr
	: expr_value 
	| expr_value neg_literal
	  {
		/* 
		** Handle '$foo = $baa-1' i.e. no space after - 
		** This isn't cought by the general purpose maths
		** rule becuase the lexer sees a negative integer
		** rather than a math_op integer pair.
		*/
		if (eCodeAlu(emberScript,(long) EMBER_OP_ADD) < 0)
			return(-1);
	  }
	| L_BRACKET expr R_BRACKET

	| expr OP_ADD expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_ADD) < 0)
			return(-1);
	  }
	| expr OP_SUB expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_SUB) < 0)
			return(-1);
	  }
	| expr OP_MUL expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_MUL) < 0)
			return(-1);
	  }
	| expr OP_DIV expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_DIV) < 0)
			return(-1);
	  }
	| expr OP_MOD expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_MOD) < 0)
			return(-1);
	  }
	| expr OP_AND expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_AND) < 0)
			return(-1);
	  }
	| expr OP_OR expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_OR) < 0)
			return(-1);
	  }
	| expr OP_SHIFT_L expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_SHIFT_L) < 0)
			return(-1);
	  }
	| expr OP_SHIFT_R expr
	  {
		if (eCodeAlu(emberScript,(long)EMBER_OP_SHIFT_R) < 0)
			return(-1);
	  }
	;

expr_value
	: literal
	| neg_literal
	| function_value
	| IDENT
	  {
		if (eCodePushMacro(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	| COUNT expr_value
	  {
		if (eCodeCount(emberScript) < 0)
			return(-1);
	  }
	| DEFINED L_BRACKET expr_value R_BRACKET
	  {
		if (eCodeDefined(emberScript) < 0)
		{
			return(-1);
		}
	  }
	| POP L_BRACKET VAR R_BRACKET
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$3);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$3);
		}
		free($3);
		if (symbol == NULL)
		{
			return(-1);
		}
		varID = symbol->id;
		if (eCodeArrayPop(emberScript, varID) < 0)
		{
			return(-1);
		}
	  }
	| SHIFT L_BRACKET VAR R_BRACKET
	  {
		e_stab	*symbol;
		int	varID;

		symbol = eSymGetVariableByName(emberScript,$3);
		if (symbol == NULL)
		{
			symbol = eSymCreateVariable(emberScript,$3);
		}
		free($3);
		if (symbol == NULL)
		{
			return(-1);
		}
		varID = symbol->id;
		if (eCodeArrayShift(emberScript, varID) < 0)
		{
			return(-1);
		}
	  }
	| VAR 
	  {
		if (eCodePushVariable(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	| VAR L_SQUARE expr R_SQUARE
	  {
		if (eCodePushVariable(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		eCodeArrayDeref(emberScript);
		free($1);
	  }
	| VAR L_BRACE expr R_BRACE
	  {
		if (eCodePushVariable(emberScript, $1) < 0)
		{
			free($1);
			return(-1);
		}
		eCodeAssocDeref(emberScript);
		free($1);
	  }
	| L_BRACKET type R_BRACKET expr_value
	  {
		if (eCodeTypeCast(emberScript, (long)$2) < 0)
			return(-1);
	  }
	| L_BRACKET type R_BRACKET L_BRACKET expr R_BRACKET
	  {
		if (eCodeTypeCast(emberScript, (long)$2) < 0)
			return(-1);
	  }
	;

function_value
	: IDENT L_BRACKET
	  {
		if (eCodeBlockIn(emberScript) < 0)
		{
			free($1);
			return(-1);
		}
	  }
	  opt_param_list R_BRACKET
	  {
		if (eCodeCall(emberScript, $1, EMBER_FALSE) < 0)
		{
			free($1);
			return(-1);
		}
		if (eCodeBlockOut(emberScript) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	;


literal
	: LIT_TEXT
	  {
		if (eCodePushLiteral(emberScript, EMBER_TYPE_TEXT, $1) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	| LIT_REAL
	  {
		if (eCodePushLiteral(emberScript, EMBER_TYPE_REAL, $1) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	| LIT_INT
	  {
		if (eCodePushLiteral(emberScript, EMBER_TYPE_INT, $1) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	;

neg_literal 
	: LIT_NEG_REAL 
	  {
		if (eCodePushLiteral(emberScript, EMBER_TYPE_REAL, $1) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	| LIT_NEG_INT
	  {
		if (eCodePushLiteral(emberScript, EMBER_TYPE_INT, $1) < 0)
		{
			free($1);
			return(-1);
		}
		free($1);
	  }
	;

cond_op 
	: COND_EQ	{ $$ = (YYSTYPE) EMBER_COND_EQ; }
	| COND_LT	{ $$ = (YYSTYPE) EMBER_COND_LT; }
	| COND_GT	{ $$ = (YYSTYPE) EMBER_COND_GT; }
	| COND_LE	{ $$ = (YYSTYPE) EMBER_COND_LE; }
	| COND_GE	{ $$ = (YYSTYPE) EMBER_COND_GE; }
	| COND_NE	{ $$ = (YYSTYPE) EMBER_COND_NE; }
	| COND_RE	{ $$ = (YYSTYPE) EMBER_COND_RE; }
	| COND_NRE	{ $$ = (YYSTYPE) EMBER_COND_NRE; }
	;

type
	: T_TEXT	{ $$ = (YYSTYPE) EMBER_TYPE_TEXT; }
	| T_ARRAY	{ $$ = (YYSTYPE) EMBER_TYPE_ARRAY; }
	| T_ASSOC	{ $$ = (YYSTYPE) EMBER_TYPE_ASSOC; }
	| T_INT		{ $$ = (YYSTYPE) EMBER_TYPE_INT; }
	| T_UINT	{ $$ = (YYSTYPE) EMBER_TYPE_UINT; }
	| T_REAL	{ $$ = (YYSTYPE) EMBER_TYPE_REAL; }
	| T_MONEY	{ $$ = (YYSTYPE) EMBER_TYPE_MONEY; }
	| T_CHAR	{ 
			  fprintf(stderr,
				"'char' is obsolete.  Use 'text' at line %d\n",
				emberScript->info.curLineNum);
			  $$ = (YYSTYPE) EMBER_TYPE_TEXT; 
		  	}
	;


html_content
	: HTML_START html_block opt_html_end
	;

html_block
	: /* NULL */
	| HTML 
	{
		e_funct *tmpFunct;
		tmpFunct = emberScript->info.curFunction;
		emberScript->info.curFunction = &emberScript->main;
		eCodeBlockIn(emberScript);
		eCodePushLiteral(emberScript, EMBER_HTML, $1);
		eCodeCall(emberScript, "echo", EMBER_TRUE);
		eCodeBlockOut(emberScript);
		emberScript->info.curFunction = tmpFunct;
	}
	html_block
	;

opt_html_end
	: /* NULL */
	| HTML_END
	;
