/*
** Copyright (c) 2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  
**
** Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: util.c,v 1.32 2006/05/20 07:01:17 bambi Exp $
**
*/

/*
** Module	: 
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/



/**************************************************************************
** STANDARD INCLUDES
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <sys/signal.h>
#include <sys/wait.h>

#include "config/config.h"
#include "ember.h"
#include "popen_exec.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif


/**************************************************************************
** MODULE SPECIFIC INCLUDES
**************************************************************************/


/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/


/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/

#define _STATE_NORMAL	0
#define	_STATE_SINGLE	1
#define _STATE_DOUBLE	2

char ** _splitCommandIntoArgv(command)
	char	*command;
{
	char	*command_dup,
		*start,
		*rc,
		*wc,
		**argv;
	int	state,
		argc;

	if (!command)
	{
		return(NULL);
	}
	command_dup = strdup(command);
	argc = 0;
	argv = NULL;
	start = rc = wc = command_dup;
	state = _STATE_NORMAL;
	while(*rc)
	{
		switch(*rc)
		{
			case '\\':
				rc++;
				if (*rc)
				{
					*wc++ = *rc++;
				}
				break;

			case ' ':
				if (state == _STATE_NORMAL)
				{
					/* skip multiple spaces */
					while(*rc == ' ')
					{
						rc++;
					}
					*wc = 0;
					argc++;
					argv = (char**)realloc(argv, argc * 
						sizeof(char**));
					argv[argc - 1] = start;
					wc++;
					start = wc;
				}
				else
				{
					*wc++ = *rc++;
				}
				break;

			case '\'':
				if (state == _STATE_NORMAL)
				{
					state = _STATE_SINGLE;
					rc++;
					break;
				}
				if (state == _STATE_SINGLE)
				{
					state = _STATE_NORMAL;
					rc++;
					break;
				}
				*wc++ = *rc++;
				break;
				
			case '\"':
				if (state == _STATE_NORMAL)
				{
					state = _STATE_DOUBLE;
					rc++;
					break;
				}
				if (state == _STATE_DOUBLE)
				{
					state = _STATE_NORMAL;
					rc++;
					break;
				}
				*wc++ = *rc++;
				break;
				
			default:
				*wc++ = *rc++;
				break;
		}
	}
	*wc = 0;
	argc++;
	argv = (char**)realloc(argv, (argc + 1) * sizeof(char**));
	argv[argc - 1] = start;
	argv[argc] = NULL;
	return(argv);
}

/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/




static int *pids;

FILE *
popen_exec(program, type)
	char *program;
	register char *type;
{
	FILE *iop;
	int pdes[2], fds, pid;
	int test_pipe[2], exec_res, res;
	char **args;


	if ((*type != 'r' && *type != 'w') || type[1])
		return(NULL);
	if (program == NULL)
		return(NULL);

	if (pids == NULL) {
		if ((fds = getdtablesize()) <= 0)
			return (NULL);
		if ((pids = (int *)malloc((u_int)(fds * sizeof(int)))) == NULL)
			return (NULL);
		bzero((char *)pids, fds * sizeof(int));
	}
	if (pipe(pdes) < 0)
		return (NULL);
	if (pipe(test_pipe) < 0)
		return (NULL);
	switch (pid = vfork()) {
	case -1:			/* error */
		(void) close(pdes[0]);
		(void) close(pdes[1]);
		(void) close(test_pipe[0]);
		(void) close(test_pipe[1]);
		return (NULL);
		/* NOTREACHED */
	case 0:				/* child */
		if (*type == 'r') {
			if (pdes[1] != fileno(stdout)) {
				(void) dup2(pdes[1], fileno(stdout));
				(void) close(pdes[1]);
			}
			(void) close(pdes[0]);
		} else {
			if (pdes[0] != fileno(stdin)) {
				(void) dup2(pdes[0], fileno(stdin));
				(void) close(pdes[0]);
			}
			(void) close(pdes[1]);
		}
		(void) close(test_pipe[0]);
		if (fcntl(test_pipe[1], F_SETFD, FD_CLOEXEC) < 0)
		{
			perror("fcntl failed");
			exit(-1);
		}
		args = _splitCommandIntoArgv(program);
		if (args == NULL)
			exit(-1);
		execv(args[0], args);
		/* exec failed.  report errno to parent */
		exec_res = errno;
		write(test_pipe[1], &exec_res, sizeof(exec_res));
		_exit(127);
		/* NOTREACHED */
	}

	close(test_pipe[1]);
	res = read(test_pipe[0], &exec_res, sizeof(exec_res));
	if (res)
	{
		/*exec in child failed */
		close(pdes[0]);
		close(pdes[1]);
		close(test_pipe[0]);
		close(test_pipe[1]);
		errno = exec_res;
		return(NULL);
	}
	/* parent; assume fdopen can't fail...  */
	if (*type == 'r') {
		iop = fdopen(pdes[0], type);
		(void) close(pdes[1]);
	} else {
		iop = fdopen(pdes[1], type);
		(void) close(pdes[0]);
	}
	pids[fileno(iop)] = pid;
	return (iop);
}

int pclose_exec(iop)
	FILE *iop;
{
	register int fdes;
	sigset_t omask, nmask;
	union wait pstat;
	register int pid;

	/*
	 * pclose returns -1 if stream is not associated with a
	 * `popened' command, if already `pclosed', or waitpid
	 * returns an error.
	 */
	if (pids == NULL || pids[fdes = fileno(iop)] == 0)
		return (-1);
	(void) fclose(iop);
	sigemptyset(&nmask);
	sigaddset(&nmask, SIGINT);
	sigaddset(&nmask, SIGQUIT);
	sigaddset(&nmask, SIGHUP);
	(void) sigprocmask(SIG_BLOCK, &nmask, &omask);
	do {
		pid = waitpid(pids[fdes], (int *) &pstat, 0);
	} while (pid == -1 && errno == EINTR);
	(void) sigprocmask(SIG_SETMASK, &omask, NULL);
	pids[fdes] = 0;
	return (pid == -1 ? -1 : pstat.w_status);
}


#ifdef DEBUG

main(argc, argv)
	int	argc;
	char	*argv[];
{
	char	**args,
		buf[1025];
	FILE	*fp;

	if (argc < 2)
	{
		printf("\n\nUsage : popen_exec 'Command to pass to popen'\n\n");
		exit(1);
	}
	fp = popen_exec(argv[1], "r");
	if (!fp)
	{
		printf("Caught popen() failure\n");
		perror("popen");
		/*strerror(errno));*/
		exit(-1);
	}
	while(fgets(buf, 1024, fp))
	{
		printf("Line : %s", buf);
	}
	pclose_exec(fp);
}

#endif
